@@ -40,6 +40,11 @@
 #include <sys/utsname.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#ifdef __linux__
+/* headers for ARP */
+#include <net/if_arp.h>
+#include <sys/ioctl.h>
+#endif
 #else
 #define _WIN32_WINNT 0x502
 #include <winsock2.h>
@@ -76,6 +81,7 @@ typedef unsigned long in_addr_t;
 #include "gssdp-socket-source.h"
 #include "gssdp-marshal.h"
 #include "gssdp-protocol.h"
+#include "gssdp-socket-functions.h"
 #ifdef HAVE_PKTINFO
 #include "gssdp-pktinfo-message.h"
 #endif
@@ -125,6 +131,7 @@ typedef struct _GSSDPHeaderField GSSDPHeaderField;
 struct _GSSDPClientPrivate {
         char              *server_id;
 
+        GHashTable        *user_agent_cache;
         guint              socket_ttl;
         guint              msearch_port;
         GSSDPNetworkDevice device;
@@ -181,6 +188,10 @@ gssdp_client_initable_init    (GInitable     *initable,
                                GCancellable  *cancellable,
                                GError       **error);
 
+char *
+arp_lookup                    (GSSDPClient   *client,
+                               const char    *ip_address);
+
 static void
 gssdp_client_init (GSSDPClient *client)
 {
@@ -317,6 +328,11 @@ gssdp_client_initable_init (GInitable                   *initable,
 
         client->priv->initialized = TRUE;
 
+        client->priv->user_agent_cache = g_hash_table_new_full (g_str_hash,
+                                                                g_str_equal,
+                                                                g_free,
+                                                                g_free);
+
         return TRUE;
 }
 
@@ -438,6 +454,11 @@ gssdp_client_dispose (GObject *object)
                 client->priv->search_socket = NULL;
         }
 
+        if (client->priv->device.host_addr != NULL) {
+                g_object_unref (client->priv->device.host_addr);
+                client->priv->device.host_addr = NULL;
+        }
+
         G_OBJECT_CLASS (gssdp_client_parent_class)->dispose (object);
 }
 
@@ -456,6 +477,9 @@ gssdp_client_finalize (GObject *object)
         g_free (client->priv->device.host_ip);
         g_free (client->priv->device.network);
 
+        if (client->priv->user_agent_cache)
+                g_hash_table_unref (client->priv->user_agent_cache);
+
         G_OBJECT_CLASS (gssdp_client_parent_class)->finalize (object);
 }
 
@@ -816,6 +840,63 @@ gssdp_client_set_network (GSSDPClient *client,
         g_object_notify (G_OBJECT (client), "network");
 }
 
+/**
+ * gssdp_client_add_cache_entry:
+ * @client: A #GSSDPClient
+ * @ip_address: The host to add to the cache
+ * @user_agent: User agent ot the host to add
+ **/
+void
+gssdp_client_add_cache_entry (GSSDPClient  *client,
+                               const char   *ip_address,
+                               const char   *user_agent)
+{
+        char *hwaddr;
+
+        g_return_if_fail (client != NULL);
+        g_return_if_fail (ip_address != NULL);
+        g_return_if_fail (user_agent != NULL);
+
+        hwaddr = arp_lookup (client, ip_address);
+
+        if (hwaddr)
+                g_hash_table_insert (client->priv->user_agent_cache,
+                                     hwaddr,
+                                     g_strdup (user_agent));
+}
+
+/**
+ * gssdp_client_guess_user_agent:
+ * @client: A #GSSDPClient
+ * @ip_address: IP address to guess the user-agent for
+ *
+ * Returns: (transfer none): The user-agent cached for this IP, %NULL if none
+ * is cached.
+ **/
+const char *
+gssdp_client_guess_user_agent (GSSDPClient *client,
+                               const char  *ip_address)
+{
+        char *hwaddr;
+
+        g_return_val_if_fail (GSSDP_IS_CLIENT (client), NULL);
+        g_return_val_if_fail (ip_address != NULL, NULL);
+
+        hwaddr = arp_lookup (client, ip_address);
+
+        if (hwaddr) {
+                const char *agent;
+
+                agent =  g_hash_table_lookup (client->priv->user_agent_cache,
+                                              hwaddr);
+                g_free (hwaddr);
+
+                return agent;
+        }
+
+        return NULL;
+}
+
 /**
  * gssdp_client_get_network:
  * @client: A #GSSDPClient
@@ -1167,7 +1248,7 @@ socket_source_cb (GSSDPSocketSource *socket_source, GSSDPClient *client)
                 goto out;
         }
 
-#ifdef HAVE_PKTINFO
+#if defined(HAVE_PKTINFO) && !defined(__APPLE__)
         {
                 int i;
                 for (i = 0; i < num_messages; i++) {
@@ -1265,7 +1346,20 @@ socket_source_cb (GSSDPSocketSource *socket_source, GSSDPClient *client)
         ip_string = g_inet_address_to_string (inetaddr);
         port = g_inet_socket_address_get_port (
                                         G_INET_SOCKET_ADDRESS (address));
+
         if (type >= 0) {
+                const char *agent;
+
+                /* update client cache */
+                agent = soup_message_headers_get_one (headers, "Server");
+                if (!agent)
+                        agent = soup_message_headers_get_one (headers, "User-Agent");
+
+                if (agent)
+                        gssdp_client_add_cache_entry (client,
+                                                      ip_string,
+                                                      agent);
+
                 g_signal_emit (client,
                                signals[MESSAGE_RECEIVED],
                                0,
@@ -1382,7 +1476,10 @@ extract_address_and_prefix (PIP_ADAPTER_UNICAST_ADDRESS  adapter,
 static int
 query_ifindex (const char *iface_name)
 {
-#ifdef HAVE_SIOCGIFINDEX
+#if defined(HAVE_IFNAMETOINDEX)
+        return if_nametoindex (iface_name);
+
+#elif defined(HAVE_SIOCGIFINDEX)
         int fd;
         int result;
         struct ifreq ifr;
@@ -1838,3 +1935,44 @@ init_network_info (GSSDPClient *client, GError **error)
         return ret;
 }
 
+char *
+arp_lookup (GSSDPClient *client, const char *ip_address)
+{
+#if defined(__linux__)
+        struct arpreq req;
+        struct sockaddr_in *sin;
+        GSocket *socket;
+
+        memset (&req, 0, sizeof (req));
+
+        /* FIXME: Update when we support IPv6 properly */
+        sin = (struct sockaddr_in *) &req.arp_pa;
+        sin->sin_family = AF_INET;
+        sin->sin_addr.s_addr = inet_addr (ip_address);
+
+        strncpy (req.arp_dev,
+                 client->priv->device.iface_name,
+                 sizeof (req.arp_dev) - 1 /* nul terminator */);
+        socket = gssdp_socket_source_get_socket (client->priv->search_socket);
+
+        if (ioctl (g_socket_get_fd (socket), SIOCGARP, (caddr_t) &req) < 0) {
+                return NULL;
+        }
+
+        if (req.arp_flags & ATF_COM) {
+                unsigned char *buf = (unsigned char *) req.arp_ha.sa_data;
+
+                return g_strdup_printf ("%02X:%02X:%02X:%02X:%02X:%02X",
+                                        buf[0],
+                                        buf[1],
+                                        buf[2],
+                                        buf[3],
+                                        buf[4],
+                                        buf[5]);
+        }
+
+        return NULL;
+#else
+        return g_strdup (ip_address);
+#endif
+}
