@@ -236,7 +236,7 @@ gssdp_socket_source_do_init (GInitable                   *initable,
                 /* UDA/1.0 says 4, UDA/1.1 says 2 */
                 self->priv->ttl = 4;
 
-        g_socket_set_multicast_ttl (self->priv->socket, 4);
+        g_socket_set_multicast_ttl (self->priv->socket, self->priv->ttl);
 
 
         /* Set up additional things according to the type of socket desired */
